/*jshint browsers:true, forin: true, laxbreak: true */
/*global test: true, start: true, stop: true, module: true, ok: true, equal: true, BrowserID:true */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
(function() {
  "use strict";

  var controller,
      el = $("body"),
      bid = BrowserID,
      storage = bid.Storage,
      network = bid.Network,
      xhr = bid.Mocks.xhr,
      emailRegistered = false,
      userCreated = true,
      mediator = bid.Mediator,
      registrations = [],
      testHelpers = bid.TestHelpers,
      register = testHelpers.register,
      provisioning = bid.Mocks.Provisioning;

  function reset() {
    emailRegistered = false;
    userCreated = true;
  }

  function createController(options) {
    options = options || {};
    controller = bid.Modules.Authenticate.create();
    controller.start(options);
  }

  module("controllers/authenticate", {
    setup: function() {
      reset();
      testHelpers.setup();
      createController();
    },

    teardown: function() {
      if (controller) {
        try {
          controller.destroy();
        } catch(e) {
          // may already be destroyed from close inside of the controller.
        }
      }
      reset();
      testHelpers.teardown();
    }
  });


  test("setting email address prefills address field", function() {
      controller.destroy();
      $("#email").val("");
      createController({ email: "registered@testuser.com" });
      equal($("#email").val(), "registered@testuser.com", "email prefilled");
  });

  function testUserUnregistered() {
    register("create_user", function() {
      ok(true, "email was valid, user not registered");
      start();
    });

    controller.checkEmail();
  }

  asyncTest("checkEmail with unknown secondary email, expect 'create_user' message", function() {
    $("#email").val("unregistered@testuser.com");
    xhr.useResult("unknown_secondary");

    testUserUnregistered();
  });

  asyncTest("checkEmail with email with leading/trailing whitespace, user not registered, expect 'create_user' message", function() {
    $("#email").val("    unregistered@testuser.com   ");
    xhr.useResult("unknown_secondary");

    testUserUnregistered();
  });

  asyncTest("checkEmail with normal email, user registered, expect 'enter_password' message", function() {
    $("#email").val("registered@testuser.com");
    xhr.useResult("known_secondary");

    register("enter_password", function() {
      ok(true, "email was valid, user registered");
      start();
    });

    controller.checkEmail();
  });

  asyncTest("checkEmail with email that has IdP support, expect 'primary_user' message", function() {
    $("#email").val("unregistered@testuser.com");
    xhr.useResult("primary");

    register("primary_user", function(msg, info) {
      equal(info.email, "unregistered@testuser.com", "email correctly passed");
      equal(info.auth, "https://auth_url", "IdP authentication URL passed");
      equal(info.prov, "https://prov_url", "IdP provisioning URL passed");
      start();
    });

    controller.checkEmail();
  });

  function testAuthenticated() {
    register("authenticated", function() {
      ok(true, "user authenticated as expected");
      start();
    });
    controller.authenticate();
  }

  asyncTest("normal authentication is kosher", function() {
    $("#email").val("registered@testuser.com");
    $("#password").val("password");

    testAuthenticated();
  });

  asyncTest("leading/trailing whitespace on the username is stripped for authentication", function() {
    $("#email").val("    registered@testuser.com    ");
    $("#password").val("password");

    testAuthenticated();
  });

  asyncTest("forgotPassword triggers forgot_password message", function() {
    $("#email").val("registered@testuser.com");

    register("forgot_password", function(msg, info) {
      equal(info.email, "registered@testuser.com", "forgot_password with correct email triggered");
      start();
    });

    controller.forgotPassword();
  });

  asyncTest("createUser with valid email", function() {
    $("#email").val("unregistered@testuser.com");
    xhr.useResult("unknown_secondary");

    register("user_staged", function(msg, info) {
      equal(info.email, "unregistered@testuser.com", "user_staged with correct email triggered");
      start();
    });

    controller.createUser();
  });

  asyncTest("createUser with invalid email", function() {
    $("#email").val("unregistered");

    var handlerCalled = false;
    register("user_staged", function(msg, info) {
      handlerCalled = true;
    });

    controller.createUser(function() {
      equal(handlerCalled, false, "bad jiji, user_staged should not have been called with invalid email");
      start();
    });
  });

  asyncTest("createUser with valid email but throttling", function() {
    $("#email").val("unregistered@testuser.com");

    var handlerCalled = false;
    register("user_staged", function(msg, info) {
      handlerCalled = true;
    });

    xhr.useResult("throttle");
    controller.createUser(function() {
      equal(handlerCalled, false, "bad jiji, user_staged should not have been called with throttling");
      equal(bid.Tooltip.shown, true, "tooltip is shown");
      start();
    });
  });

  asyncTest("createUser with valid email, XHR error", function() {
    $("#email").val("unregistered@testuser.com");

    var handlerCalled = false;
    register("user_staged", function(msg, info) {
      handlerCalled = true;
    });

    xhr.useResult("ajaxError");
    controller.createUser(function() {
      equal(handlerCalled, false, "bad jiji, user_staged should not have been called with XHR error");
      start();
    });
  });

}());

